# PAPI LDAPAuth functions
# Copyright (c) 2001-2005, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.
#
# These functions implement PAPI user authentication and credential data 
# provision using LDAP.
#
# The schema used by these functions is included in the file "papi.schema"
#

package PAPI::LDAPAuth;
use strict;
use Net::LDAP;
use Net::LDAP::Message;
use Net::LDAP::Search;
use Net::LDAP::Entry;
use Digest::MD5 qw(md5_base64);

# This function is intended to be hooked at the authenticationHook in a
# PAPI AuthServer. It uses the connection variable 'password' (and any
# other required by the DN template, see below) to validate an
# authentication request against the LDAP server defined by the
# configuration variables 'LDAPserver' (defaults to "ldap") and
# 'LDAPport' (defaults to 389).
# If the configuration variable 'LDAPS' is set to 1, the connection will
# use TLS (and the default port will be 636). TLS parameters can configured
# by means of 'LDAPS*' variables.
# Authentication is performed by binding to the LDAP server using a
# specific DN and the value of 'password'. The DN is established by searching
# from 'LDAPAuthSearchBase' with the filter built from the 'LDAPUSERtemplate'
# variable and the 'objectClass=papiUser'. Search is performed using the
# scope defined by 'LDAPAuthScope' (subtree by default).
# Connection and/or configuration variables can be substituted into 
# 'LDAPUSERtemplate' and 'LDAPAuthSearchBase' using the syntax:
# <papi var="VARNAME"/>, the same used for substituting PAPI AuthServer
# variables in other templates. By default, the template is
# 'uid=<papi var="username"/>'
# The initial bind intended to establish user's DN is made anonymously
# unless the configuration variables 'LDAPbindDN' and 'LDAPbindPassword'
# are configured
#
sub VerifyUser {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $template = 'uid=<papi var="username"/>';
   my $baseuser = "";
   my $rtu = "LDAPuser";
   my $inibdn = undef;
   my $inibpw = undef;
   my $usc = "sub";
   my $ssl = undef;
   my @ssld = ();
   my ($k, $ures, $ufil, $entry, $uen, $acc);
   
   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $template = $$cfg{'LDAPUSERtemplate'} if (exists $$cfg{'LDAPUSERtemplate'});
   $baseuser = $$cfg{'LDAPAuthSearchBase'} if (exists $$cfg{'LDAPAuthSearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $usc = $$cfg{'LDAPAuthScope'} if (exists $$cfg{'LDAPAuthScope'});
   $rtu = $$con{'username'} if (exists $$con{'username'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return undef;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return undef;
   }
   $ufil="(\&(objectClass=papiUser)($template))";
 
   $ures = $ldapc->search(base => $baseuser, filter => $ufil, scope => $usc);
   if ($ures->error ne "Success" || !defined ($ures->entry(0))) {
      $$con{'PAPIerror'} = "LDAP entry not found: ".$ures->error;
      $ldapc->unbind ();
      return undef;
   }

   $entry=$ures->entry(0);
   $uen=$entry->dn;

   $acc=$ldapc->bind($uen,password => $$con{'password'});
   if ($acc->error ne "Success"){
      $$con{'PAPIerror'} = "Failed LDAP bind: ".$acc->error;
      return undef;
   }

   $ldapc->unbind ();
   return $rtu;
}

# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It uses LDAP to retrieve the list of site definitions a user
# can connect to. It returns data for the sites included in the user entry
# itself and in the entries for those groups the user belongs to. 
# The user definition into the LDAP server is located using a DN built
# from the configuration variable 'LDAPUSERtemplate'. Connection and/or
# configuration variables are substituted into the DN template and the
# search bases using the syntax: <papi var="VARNAME"/>, the same used for
# substituting PAPI AuthServer variables in other templates. By default,
# the template is 'uid=<papi var="PAPIuid"/>'.
# The function makes a bind (anonymous if the configuration variables
# 'LDAPbindDN' and 'LDAPbindPassword' are not defined)to the LDAP server
# defined by the configuration variables 'LDAPserver' (defaults to "ldap") and
# 'LDAPport' (defaults to 389).
# If the configuration variable 'LDAPS' is set to 1, the connection will
# use TLS (and the default port will be 636). TLS parameters can configured
# by means of 'LDAPS*' variables.
# Objects are searched in the server using the LDAP search base defined by
# the configuration variable 'LDAPsearchBase' (defaults to "").
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
# 
sub UserCredentials {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $baseuser = "";
   my $template = 'uid=<papi var="PAPIuid"/>';
   my $inibdn = undef;
   my $inibpw = undef;
   my $usc = "sub";
   my $ssl = undef;
   my @ssld = ();
   my @retar = ();
   my @sites = ();
   my @sitelist = ();
   my @grouplist = ();
   my @genlist = ();
   my ($k, $ufil, $ures, $uen, $gfil, $gres, $gen,
       $sfil, $sres, $sen, $six, $assertion, $entry);

   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $baseuser = $$cfg{'LDAPAuthSearchBase'} if (exists $$cfg{'LDAPAuthSearchBase'});
   $template = $$cfg{'LDAPUSERtemplate'} if (exists $$cfg{'LDAPUSERtemplate'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $usc = $$cfg{'LDAPAuthScope'} if (exists $$cfg{'LDAPAuthScope'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return \@retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return \@retar;
   }
#
# Get the user's entry
   $ufil="(\&(objectClass=papiUser)($template))";
   $ures = $ldapc->search(base => $baseuser, filter => $ufil, scope => $usc);
   if ($ures->error ne "Success" || !defined ($ures->entry(0))) {
      $$con{'PAPIerror'} = "LDAP entry not found: ".$ures->error;
      $ldapc->unbind ();
      return \@retar;
   }
   $uen = $ures->entry(0);
#
# Build a list of site identifiers, using first user's own references and
# then group references. Remove duplicate identifiers. Set 'PAPIgid'.
   @sitelist=($uen->get_value('papiSiteId')) if ($uen->exists('papiSiteId'));
   @grouplist=($uen->get_value('papiGroupId')) if ($uen->exists('papiGroupId'));
   for my $group (@grouplist) {
      $gfil = "(\&(papiGroupId=$group)(objectClass=papiGroup))";
      $gres = $ldapc->search (base => $base, filter => $gfil);
      if (defined $gres->entry(0)) {
         $gen = $gres->entry(0);
         push @genlist, $gen;
         push @sitelist, $gen->get_value ('papiSiteId')
              if ($gen->exists ('papiSiteId'));
         if ($gen->exists ('description') &&
             (!exists $$con{'PAPIgid'} || $$con{'PAPIgid'} eq "")) {
            $$con{'PAPIgid'} = $gen->get_value ('description');
         }
      }
   }
   for my $tsite (@sitelist) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Build the list of hashes from the site data stored into the LDAP server
   for my $site (@sites) {
      my $sitedef = {};
      $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
      $sres = $ldapc->search (base => $base, filter => $sfil);
      if (defined $sres->entry(0)) {     
         $sen = $sres->entry(0);
         $sitedef->{'desc'} = $sen->get_value ('description')
                              if ($sen->exists ('description'));
         $sitedef->{'poa'} = $sen->get_value ('papiSitePoA')
                              if ($sen->exists ('papiSitePoA'));
         $sitedef->{'authURI'} = $sen->get_value ('papiSiteAuth')
                              if ($sen->exists ('papiSiteAuth'));
         $sitedef->{'accessURI'} = $sen->get_value ('papiSiteAccess')
                              if ($sen->exists ('papiSiteAccess'));
         $sitedef->{'ttl'} = $sen->get_value ('papiSiteTtl')
                              if ($sen->exists ('papiSiteTtl'));
         $sitedef->{'service'} = $sen->get_value ('papiSiteService')
                              if ($sen->exists ('papiSiteService'));
         $sitedef->{'location'} = $sen->get_value ('papiSiteLocation')
                              if ($sen->exists ('papiSiteLocation'));
         $sitedef->{'acceptURL'} = $sen->get_value ('papiSiteAcceptURL')
                              if ($sen->exists ('papiSiteAcceptURL'));
         $sitedef->{'rejectURL'} = $sen->get_value ('papiSiteRejectURL')
                              if ($sen->exists ('papiSiteRejectURL'));
         $assertion = buildAssertion ($sen, \@genlist, $uen);
         $sitedef->{'assertion'} = $assertion if (defined $assertion);
         push @retar, $sitedef;
      }
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   $ldapc->unbind ();
   return \@retar;
}

# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It returns a site list consisting of the sites related to
# the groups whose identifiers are included into the configuration
# variable 'groupList'. The list has to be comma-separated.
# The function makes a bind (anonymous if the configuration variables
# 'LDAPbindDN' and 'LDAPbindPassword' are not defined) to the LDAP server
# defined by the configuration variables 'LDAPserver' (defaults to "ldap") and
# 'LDAPport' (defaults to 389).
# If the configuration variable 'LDAPS' is set to 1, the connection will
# use TLS (and the default port will be 636). TLS parameters can configured
# by means of 'LDAPS*' variables.
# Objects are searched in the server using the LDAP search base defined by
# the configuration variable 'LDAPsearchBase' (defaults to "").
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
#
sub GroupList {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my @retar = ();
   my @genlist = ();
   my @sitelist = ();
   my @sites = ();
   my ($gfil, $gres, $gen, $sfil, $sres, $sen, $six, $assertion, $k);

   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return \@retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }

   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return \@retar;
   }
#
# Build a list of site identifiers. Remove duplicate identifiers. Set 'PAPIgid'.
   for my $group (split /\s*,\s*/, $$cfg{'groupList'}) {
      $gfil = "(\&(papiGroupId=$group)(objectClass=papiGroup))";
      $gres = $ldapc->search (base => $base, filter => $gfil);
      if (defined $gres->entry(0)) {
         $gen = $gres->entry(0);
         push @genlist, $gen;
         push @sitelist, $gen->get_value ('papiSiteId')
              if ($gen->exists ('papiSiteId'));
         if ($gen->exists ('description') &&
             (!exists $$con{'PAPIgid'} || $$con{'PAPIgid'} eq "")) {
            $$con{'PAPIgid'} = $gen->get_value ('description');
         }
      }
   }
   for my $tsite (@sitelist) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
   #
# Build the list of hashes from the site data stored into the database
   for my $site (@sites) {
      my $sitedef = {};
      $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
      $sres = $ldapc->search (base => $base, filter => $sfil);
      if (defined $sres->entry(0)) {
         $sen = $sres->entry(0);
         $sitedef->{'desc'} = $sen->get_value ('description')
                              if ($sen->exists ('description'));
         $sitedef->{'poa'} = $sen->get_value ('papiSitePoA')
                              if ($sen->exists ('papiSitePoA'));
         $sitedef->{'authURI'} = $sen->get_value ('papiSiteAuth')
                              if ($sen->exists ('papiSiteAuth'));
         $sitedef->{'accessURI'} = $sen->get_value ('papiSiteAccess')
                              if ($sen->exists ('papiSiteAccess'));
         $sitedef->{'ttl'} = $sen->get_value ('papiSiteTtl')
                              if ($sen->exists ('papiSiteTtl'));
         $sitedef->{'service'} = $sen->get_value ('papiSiteService')
                              if ($sen->exists ('papiSiteService'));
         $sitedef->{'location'} = $sen->get_value ('papiSiteLocation')
                              if ($sen->exists ('papiSiteLocation'));
         $sitedef->{'acceptURL'} = $sen->get_value ('papiSiteAcceptURL')
                              if ($sen->exists ('papiSiteAcceptURL'));
         $sitedef->{'rejectURL'} = $sen->get_value ('papiSiteRejectURL')
                              if ($sen->exists ('papiSiteRejectURL'));
         $assertion = buildAssertion ($sen, \@genlist);
         $sitedef->{'assertion'} = $assertion if (defined $assertion);
         push @retar, $sitedef;
      }
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   $ldapc->unbind ();
   return \@retar;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It returns a site list consisting of all the sites known
# to the LDAP server.
# The function makes a bind (anonymous if the configuration variables
# 'LDAPbindDN' and 'LDAPbindPassword' are not defined) to the LDAP server
# defined by the configuration variables 'LDAPserver' (defaults to "ldap") and
# 'LDAPport' (defaults to 389).
# If the configuration variable 'LDAPS' is set to 1, the connection will
# use TLS (and the default port will be 636). TLS parameters can configured
# by means of 'LDAPS*' variables.
# Objects are searched in the server using the LDAP search base defined by
# the configuration variable 'LDAPsearchBase' (defaults to "").
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
#
sub AllSites {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my @retar = ();
   my @sites = ();
   my ($sfil, $sres, $sen, $i, $assertion, $k);

   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return \@retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }

   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return \@retar;
   }
#
# Search the LDAP server for object of class 'papiSite'
   $sfil = "(objectClass=papiSite)";
   $sres = $ldapc->search (base => $base, filter => $sfil);
   for ($i = 0; $i < $sres->count; $i++) {
      my $sitedef = {};
      $sen = $sres->entry($i);
      $sitedef->{'desc'} = $sen->get_value ('description')
                           if ($sen->exists ('description'));
      $sitedef->{'poa'} = $sen->get_value ('papiSitePoA')
                           if ($sen->exists ('papiSitePoA'));
      $sitedef->{'authURI'} = $sen->get_value ('papiSiteAuth')
                           if ($sen->exists ('papiSiteAuth'));
      $sitedef->{'accessURI'} = $sen->get_value ('papiSiteAccess')
                            if ($sen->exists ('papiSiteAccess'));
      $sitedef->{'ttl'} = $sen->get_value ('papiSiteTtl')
                           if ($sen->exists ('papiSiteTtl'));
      $sitedef->{'service'} = $sen->get_value ('papiSiteService')
                           if ($sen->exists ('papiSiteService'));
      $sitedef->{'location'} = $sen->get_value ('papiSiteLocation')
                           if ($sen->exists ('papiSiteLocation'));
      $sitedef->{'acceptURL'} = $sen->get_value ('papiSiteAcceptURL')
                           if ($sen->exists ('papiSiteAcceptURL'));
      $sitedef->{'rejectURL'} = $sen->get_value ('papiSiteRejectURL')
                           if ($sen->exists ('papiSiteRejectURL'));
      $assertion = buildAssertion ($sen);
      $sitedef->{'assertion'} = $assertion if (defined $assertion);
      push @retar, $sitedef;
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   $ldapc->unbind ();
   return \@retar;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It just returns an empty site list, so the default values
# configured for the AuthServer are used. To be used as a stub when no other
# more sophisticated credentials assignment methods are requested.
#
sub DefCredentials {
   my @retar = ();
   push @retar, {};
   return \@retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the LDAP base defined by the configuration
# variables 'LDAPserver', 'LDAPport', 'LDAPS*' and 'LDAPsearchBase'.
# User data is determined by means of the DN retrieved using 'LDAPUSERtemplate'.
# If a suitable site is found, it returns the applicable assertion format
# and the time-to-live defined for the site. If no site is found matching
# the requesting URL, it sets the PAPIerror variables and a negative value
# is returned.
sub UserAttributes {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $template = 'uid=<papi var="PAPIuid"/>';
   my $baseuser = "";
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my @retar = (-1, undef);
   my @sites = ();
   my @sitelist = ();
   my @grouplist = ();
   my @genlist = ();
   my ($k, $ufil, $ures, $uen, $gfil, $gres, $gen,
       $sfil, $sres, $sen, $six);
 
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $template = $$cfg{'LDAPUSERtemplate'} if (exists $$cfg{'LDAPUSERtemplate'});
   $baseuser = $$cfg{'LDAPAuthSearchBase'} if (exists $$cfg{'LDAPAuthSearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return @retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return @retar;
   }
#
# Get the user's entry
   $ufil = "(\&(objectClass=papiUser)($template))";
   $ures = $ldapc->search(base => $baseuser, filter => $ufil);
   if ($ures->error ne "Success" || !defined ($ures->entry(0))) {
      $$con{'PAPIerror'} = "LDAP entry not found: ".$ures->error;
      $ldapc->unbind ();
      return @retar;
   }
   $uen = $ures->entry(0);
#
# Build a list of site identifiers, using first user's own references and
# then group references. Remove duplicate identifiers.
   @sitelist=($uen->get_value('papiSiteId')) if ($uen->exists('papiSiteId'));
   @grouplist=($uen->get_value('papiGroupId')) if ($uen->exists('papiGroupId'));   for my $group (@grouplist) {
      $gfil = "(\&(papiGroupId=$group)(objectClass=papiGroup))";
      $gres = $ldapc->search (base => $base, filter => $gfil);
      if (defined $gres->entry(0)) {
         $gen = $gres->entry(0);
         push @genlist, $gen;
         push @sitelist, $gen->get_value ('papiSiteId')
              if ($gen->exists ('papiSiteId'));
      }
   }
   for my $tsite (@sitelist) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Explore the sites in the site list. Keep the longest match
   my $len = 0;
   my $thesite = undef;
   my $theref = undef;
   if (defined $opoa) {
     for my $site (@sites) {
        $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
        $sres = $ldapc->search (base => $base, filter => $sfil);
        if (defined $sres->entry(0)) {
           $sen = $sres->entry(0);
           my $siu = $sen->get_value('papiSitePoA').$sen->get_value ('papiSiteLocation');
           if ($opoa =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = $site;
           }
        }
     }
   }
   if (!defined $thesite) {
     for my $site (@sites) {
        $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
        $sres = $ldapc->search (base => $base, filter => $sfil);
        if (defined $sres->entry(0)) {
           $sen = $sres->entry(0);
           my $siu = $sen->get_value('papiSitePoA').$sen->get_value ('papiSiteLocation');
           if ($poaurl =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = $site;
           }
        }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $theref = undef; $thesite = "__PAPIdefault__";
      }
   }
   if (!defined $thesite) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      $ldapc->unbind ();
      return @retar;
   }
   if (defined $theref && $theref->exists ('papiSiteTtl')) {
      $retar[0] = $theref->get_value ('papiSiteTtl');
   }
   else { $retar[0] = $$cfg{'defTimeToLive'}; }
   if (defined $theref) {
      $retar[1] = buildAssertion ($theref, \@genlist, $uen);
   }
   else { $retar[1] = expandAssertion ($uen, $$cfg{'defAssertion'}); }
   $ldapc->unbind ();
   return @retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the LDAP base defined by the configuration
# variables 'LDAPserver', 'LDAPport', 'LDAPS*' and 'LDAPsearchBase'.
# The site must be defined inside the list of papiSite objects defined by the
# configuration variable 'groupList'. If a suitable site is found, it
# returns the applicable assertion format and the time-to-live defined for
# the site. If no site is found matching the requesting URL, it sets the
# 'PAPIerror' variable and a negative value is returned.
sub GroupListAttributes {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   my @retar = (-1, undef);
   my @genlist = ();
   my @sitelist = ();
   my @sites = ();
   my ($gfil, $gres, $gen, $sfil, $sres, $sen, $six, $k);
 
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return @retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return @retar;
   }
#
# Build a list of site identifiers. Remove duplicate identifiers.
   for my $group (split /\s*,\s*/, $$cfg{'groupList'}) {
      $gfil = "(\&(papiGroupId=$group)(objectClass=papiGroup))";
      $gres = $ldapc->search (base => $base, filter => $gfil);
      if (defined $gres->entry(0)) {
         $gen = $gres->entry(0);
         push @genlist, $gen;
         push @sitelist, $gen->get_value ('papiSiteId')
              if ($gen->exists ('papiSiteId'));
      }
   }
   for my $tsite (@sitelist) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Explore the sites in the site list. Keep the longest match
   my $len = 0;
   my $thesite = undef;
   my $theref = undef;
   if (defined $opoa) {
     for my $site (@sites) {
        $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
        $sres = $ldapc->search (base => $base, filter => $sfil);
        if (defined $sres->entry(0)) {
           $sen = $sres->entry(0);
           my $siu = $sen->get_value('papiSitePoA').$sen->get_value ('papiSiteLocation');
           if ($opoa =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = $site;
           }
        }
     }
   }
   if (!defined $thesite) {
     for my $site (@sites) {
        $sfil = "(\&(papiSiteId=$site)(objectClass=papiSite))";
        $sres = $ldapc->search (base => $base, filter => $sfil);
        if (defined $sres->entry(0)) {
           $sen = $sres->entry(0);
           my $siu = $sen->get_value('papiSitePoA').$sen->get_value ('papiSiteLocation');
           if ($poaurl =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = $site;
           }
        }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $theref = undef; $thesite = "__PAPIdefault__";
      }
   }
   if (!defined $thesite) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      $ldapc->unbind ();
      return @retar;
   }
   if (defined $theref && $theref->exists ('papiSiteTtl')) {
      $retar[0] = $theref->get_value ('papiSiteTtl');
   }
   else { $retar[0] = $$cfg{'defTimeToLive'}; }
   if (defined $theref) {
      $retar[1] = buildAssertion ($theref, \@genlist);
   }
   else { $retar[1] = expandAssertion (undef, $$cfg{'defAssertion'}); }
   $ldapc->unbind ();
   return @retar;
}



# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the LDAP base defined by the configuration
# variables 'LDAPserver', 'LDAPport', 'LDAPS*' and 'LDAPsearchBase'.
# User data is only used to establish the contents of the assertions
# to be sent in response, and NOT to find the suitable site. Instead, ALL 
# defined sites are used to look for a match.
# User data is determined by means of the DN retrieved using 'LDAPUSERtemplate'
# If a suitable site is found, it returns the applicable assertion format
# and the time-to-live defined for the site. If no site is found matching
# the requesting URL, it sets the PAPIerror variables and a negative value
# is returned.
sub AnySiteAttributes {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $template = 'uid=<papi var="PAPIuid"/>';
   my $baseuser = "";
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};   
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my @retar = (-1, undef);
   my ($i, $k, $ufil, $ures, $uen, $sfil, $sres, $sen);
 
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $template = $$cfg{'LDAPUSERtemplate'} if (exists $$cfg{'LDAPUSERtemplate'});
   $baseuser = $$cfg{'LDAPAuthSearchBase'} if (exists $$cfg{'LDAPAuthSearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return @retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $template =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $baseuser =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return @retar;
   }
#
# Get the user's entry
   $ufil = "(\&(objectClass=papiUser)($template))";
   $ures = $ldapc->search(base => $baseuser, filter => $ufil);
   if ($ures->error ne "Success" || !defined ($ures->entry(0))) {
      $$con{'PAPIerror'} = "LDAP entry not found: ".$ures->error;
      $ldapc->unbind ();
      return @retar;
   }
   $uen = $ures->entry(0);
#
# Search the LDAP server for object of class 'papiSite'
   my $len = 0;
   my $thesite = undef;
   my $theref = undef;
   $sfil = "(objectClass=papiSite)";
   $sres = $ldapc->search (base => $base, filter => $sfil);
   if (defined $opoa) {
     for ($i = 0; $i < $sres->count; $i++) {
        $sen = $sres->entry($i);
        my $siu = $sen->get_value('papiSitePoA').$sen->get_value('papiSiteLocation');
        if ($opoa =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = "";
        }
     }
   }
   if (!defined $thesite) {
     for ($i = 0; $i < $sres->count; $i++) {
        $sen = $sres->entry($i);
        my $siu = $sen->get_value('papiSitePoA').$sen->get_value('papiSiteLocation');
        if ($poaurl =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = "";
        }
     }   
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $theref = undef; $thesite = "__PAPIdefault__";
      }
   }   
   if (!defined $thesite) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      $ldapc->unbind ();
      return @retar;
   }  
   if (defined $theref && $theref->exists ('papiSiteTtl')) {
      $retar[0] = $theref->get_value ('papiSiteTtl');
   }
   else { $retar[0] = $$cfg{'defTimeToLive'}; }
   if (defined $theref) {
      $retar[1] = buildAssertion ($theref, undef, $uen);
   }
   else { $retar[1] = expandAssertion ($uen, $$cfg{'defAssertion'}); }
   $ldapc->unbind ();
   return @retar;
}



# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a site defined site within the LDAP base defined by the configuration
# variables 'LDAPserver', 'LDAPport', 'LDAPS*' and 'LDAPsearchBase'.
# If a suitable site is found, it returns the applicable assertion format
# and the time-to-live defined for the site. If no site is found matching
# the requesting URL, it sets the 'PAPIerror' variable and a negative value
# is returned.
sub AllSiteAttributes {
   my $con = shift;
   my $cfg = shift;
   my $srv = "ldap";
   my $port = 389;
   my $ldapc = undef;
   my $base = "";
   my $inibdn = undef;
   my $inibpw = undef;
   my $ssl = undef;
   my @ssld = ();
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};      
   my @retar = (-1, undef);
   my @sites = ();
   my ($sfil, $sres, $sen, $i, $k);
 
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   $srv = $$cfg{'LDAPserver'} if (exists $$cfg{'LDAPserver'});
   $port = $$cfg{'LDAPport'} if (exists $$cfg{'LDAPport'});
   $base = $$cfg{'LDAPsearchBase'} if (exists $$cfg{'LDAPsearchBase'});
   $inibdn = $$cfg{'LDAPbindDN'} if (exists $$cfg{'LDAPbindDN'});
   $inibpw = $$cfg{'LDAPbindPassword'} if (exists $$cfg{'LDAPbindPassword'});
   $ssl = $$cfg{'LDAPS'} if (exists $$cfg{'LDAPS'});
   if ($ssl) {
      $port = 636 if (!exists $$cfg{'LDAPport'});
      for my $ck (keys %{$cfg}) {
         if ($ck =~ m/^LDAPS(.+?)/) { push @ssld,($1=>$$cfg{$ck}); }
      }
      my $lsurl = "ldaps://$srv:$port/";
      $ldapc = Net::LDAP->new($lsurl,@ssld);
   }
   else { $ldapc = Net::LDAP->new($srv, port => $port); }
   if (!defined $ldapc) {
      $$con{'PAPIerror'} = "Cannot connect to LDAP server";
      return @retar;
   }
#
# Substitute connection and configuration variables
   for $k (keys %$con) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$con{$k}%g;
   }
   for $k (keys %$cfg) {
      $base =~ s%<papi\s+var\s*\=\s*\"$k\"\s*/>%$$cfg{$k}%g;
   }
   my $bindr;
   if (defined $inibdn) { $bindr = $ldapc->bind($inibdn, password => $inibpw); }
   else { $bindr = $ldapc->bind (anonymous => 1); }
   if ($bindr->error ne "Success") {
      $$con{'PAPIerror'} = "Failed initial LDAP bind: ".$bindr->error;
      return @retar;
   }
#
# Search the LDAP server for object of class 'papiSite'
   my $len = 0;
   my $thesite = undef;
   my $theref = undef;
   $sfil = "(objectClass=papiSite)";
   $sres = $ldapc->search (base => $base, filter => $sfil);
   if (defined $opoa) {
     for ($i = 0; $i < $sres->count; $i++) {
        $sen = $sres->entry($i);
        my $siu = $sen->get_value('papiSitePoA').$sen->get_value('papiSiteLocation');
        if ($opoa =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = "";
        }
     }
   }
   if (!defined $thesite) {
    for ($i = 0; $i < $sres->count; $i++) {
        $sen = $sres->entry($i);
        my $siu = $sen->get_value('papiSitePoA').$sen->get_value('papiSiteLocation');
        if ($poaurl =~ m(^$siu) && length $siu > $len) {
              $theref = $sen; $thesite = "";
        }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $theref = undef; $thesite = "__PAPIdefault__";
      }
   }
   if (!defined $thesite) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      $ldapc->unbind ();
      return @retar;
   }
   if (defined $theref && $theref->exists ('papiSiteTtl')) {
      $retar[0] = $theref->get_value ('papiSiteTtl');
   }
   else { $retar[0] = $$cfg{'defTimeToLive'}; }
   if (defined $theref) { $retar[1] = buildAssertion ($theref); }
   else { $retar[1] = expandAssertion (undef, $$cfg{'defAssertion'}); }
   $ldapc->unbind ();
   return @retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to match
# against the default site defined for the AuthServer. To be used as a stub
# when no other more sophisticated methods are requested.
#
sub DefAttributes {
   my $con = shift;
   my $cfg = shift;
   my @retar = (-1, undef);
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};      
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
   if ($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) {
      $retar[0] = $$cfg{'defTimeToLive'};
      $retar[1] = expandAssertion (undef, $$cfg{'defAssertion'});
   }
   else {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";   }
   return @retar;
}


# This is an auxiliary function for calculating the correct assertion to
# be sent to the PoA defined for a given papiSite object.
# It uses the attribute papiAssertion in the papiSite object and/or the
# attributes papiQualifiedAssertion in the papiUser and/or papiGroup object(s)
# applicable.
# If a papiUser object is available, it also perform the substitutions for
# attribute values inside the selected assertion
#
sub buildAssertion {
   my $site = shift;
   my $siteId = $site->get_value ('papiSiteId');
   my $grouplistr = shift;
   my @grouplist = ();
   if (defined $grouplistr) { @grouplist = @{$grouplistr}; }
   my $user = shift;

   if (defined $user) {
      if ($user->exists('papiQualifiedAssertion')) {
         for my $qa ($user->get_value('papiQualifiedAssertion')) {
            my ($sid, $uqa) = split /[\s]+/, $qa, 2;
            if ($sid eq $siteId) { return expandAssertion ($user, $uqa); }
         }
      }
   }
   for my $group (@grouplist) {
      if ($group->exists('papiQualifiedAssertion')) {
         for my $qa ($group->get_value('papiQualifiedAssertion')) {
            my ($sid, $uqa) = split /[\s]+/, $qa, 2;
            if ($sid eq $siteId) { return expandAssertion ($user, $uqa); }
         }
      }
   }
   if ($site->exists('papiAssertion')) {
      return expandAssertion ($user, $site->get_value('papiAssertion'));
   }
   return undef;
}


# This is an auxiliary function for calculating the assertion to
# be sent to the PoA defined for a given papiSite object.
# It expands the references to user's entry attributes (in the forms of
# <papi attr="NAME"/> or <papi attrlist="NAME" sep="STRING"/>) into
# the corresponding values. If no attributes are available or the user
# entry is not defined, the corresponding references are substituted
# by an empty string. Syntax errors are also (silently) substituted,
# when detected, by the empty string.
#
sub expandAssertion {
   my $user = shift;
   my $assertion = shift;
   if (defined $user) {
      my @attrs = ($assertion =~ m/<papi\s+attr\s*\=\s*(.*?)\s*\/>/g);
      for my $at (@attrs) {
      	 my ($atv) = ($at =~ m/^\"(.*?)\"/g);
         if ($user->exists($atv)) {
            my $av = $user->get_value($atv);
            $assertion =~ s%<papi\s+attr\s*\=\s*\"$atv\"(.*?)/>%$av%g;
         }
      }
      @attrs = ($assertion =~ m/<papi\s+hashedattr\s*\=\s*(.*?)\s*\/>/g);
      for my $at (@attrs) {
      	 my ($atv) = ($at =~ m/^\"(.*?)\"/g);
         if ($user->exists($atv)) {
            my $av = $user->get_value($atv);
            $av = md5_base64($av);
            $assertion =~ s%<papi\s+hashedattr\s*\=\s*\"$atv\"(.*?)/>%$av%g;
         }
      }      
      @attrs = ($assertion =~ m/<papi\s+attrlist\s*\=\s*(.*?)\s*\/>/g);
      for my $at (@attrs) {
         my ($atn, $ats) = ($at =~ m/^\"(.*?)\"\s+sep\s*\=\s*\"(.*?)\"/g);
         if (defined $atn && defined $ats && $user->exists($atn)) {
            my @avs = $user->get_value($atn);
            my $av = join $ats,@avs;
            $assertion =~ s%<papi\s+attrlist\s*\=\s*\"$atn\"\s+sep\s*\=\s*\"$ats\"\s*/>%$av%g;
         }
      }      
   }
   my @attds = ($assertion =~ m/<papi\s+attr.*?\=.*?default\s*\=\"(.*?)\".*?\/>/g);
   for my $at (@attds) {
      $assertion =~ s%<papi\s+attr.*?\=.*?default\s*\=\"$at\"(.*?)/>%$at%g;
   }
   $assertion =~ s%<papi\s+attr.*?\=.*?/>%%g;
   return $assertion
}


1;
